import { autocompletion, insertCompletionText } from "@codemirror/autocomplete";
import { setDiagnostics } from "@codemirror/lint";
import { Facet } from "@codemirror/state";
import { EditorView, ViewPlugin, hoverTooltip, } from "@codemirror/view";
import { Client, RequestManager, WebSocketTransport, } from "@open-rpc/client-js";
import { CompletionItemKind, CompletionTriggerKind, DiagnosticSeverity, } from "vscode-languageserver-protocol";
import { formatContents, isLSPTextEdit, offsetToPos, posToOffset, posToOffsetOrZero, prefixMatch, showErrorMessage, } from "./utils.js";
const timeout = 10000;
const changesDelay = 500;
const CompletionItemKindMap = Object.fromEntries(Object.entries(CompletionItemKind).map(([key, value]) => [value, key]));
// biome-ignore lint/style/noNonNullAssertion: <explanation>
const useLast = (values) => values.at(-1);
const client = Facet.define({
    combine: useLast,
});
const documentUri = Facet.define({ combine: useLast });
const languageId = Facet.define({ combine: useLast });
const logger = console.log;
export class LanguageServerClient {
    ready;
    capabilities;
    initializePromise;
    rootUri;
    workspaceFolders;
    autoClose;
    transport;
    requestManager;
    client;
    initializationOptions;
    clientCapabilities;
    plugins;
    constructor(options) {
        this.ready = false;
        this.capabilities = null;
        this.rootUri = options.rootUri;
        this.workspaceFolders = options.workspaceFolders;
        this.autoClose = options.autoClose;
        this.plugins = [];
        this.transport = options.transport;
        this.initializationOptions = options.initializationOptions;
        this.clientCapabilities = options.capabilities;
        this.requestManager = new RequestManager([this.transport]);
        this.client = new Client(this.requestManager);
        this.client.onNotification((data) => {
            this.processNotification(data);
        });
        const webSocketTransport = this.transport;
        if (webSocketTransport?.connection) {
            // XXX(hjr265): Need a better way to do this. Relevant issue:
            // https://github.com/FurqanSoftware/codemirror-languageserver/issues/9
            webSocketTransport.connection.addEventListener("message", (message) => {
                const data = JSON.parse(message.data);
                if (data.method && data.id) {
                    webSocketTransport.connection.send(JSON.stringify({
                        jsonrpc: "2.0",
                        id: data.id,
                        result: null,
                    }));
                }
            });
        }
        this.initializePromise = this.initialize();
    }
    getInitializationOptions() {
        const defaultClientCapabilities = {
            textDocument: {
                hover: {
                    dynamicRegistration: true,
                    contentFormat: ["markdown", "plaintext"],
                },
                moniker: {},
                synchronization: {
                    dynamicRegistration: true,
                    willSave: false,
                    didSave: false,
                    willSaveWaitUntil: false,
                },
                codeAction: {
                    dynamicRegistration: true,
                    codeActionLiteralSupport: {
                        codeActionKind: {
                            valueSet: [
                                "",
                                "quickfix",
                                "refactor",
                                "refactor.extract",
                                "refactor.inline",
                                "refactor.rewrite",
                                "source",
                                "source.organizeImports",
                            ],
                        },
                    },
                    resolveSupport: {
                        properties: ["edit"],
                    },
                },
                completion: {
                    dynamicRegistration: true,
                    completionItem: {
                        snippetSupport: false,
                        commitCharactersSupport: true,
                        documentationFormat: ["markdown", "plaintext"],
                        deprecatedSupport: false,
                        preselectSupport: false,
                    },
                    contextSupport: false,
                },
                signatureHelp: {
                    dynamicRegistration: true,
                    signatureInformation: {
                        documentationFormat: ["markdown", "plaintext"],
                    },
                },
                declaration: {
                    dynamicRegistration: true,
                    linkSupport: true,
                },
                definition: {
                    dynamicRegistration: true,
                    linkSupport: true,
                },
                typeDefinition: {
                    dynamicRegistration: true,
                    linkSupport: true,
                },
                implementation: {
                    dynamicRegistration: true,
                    linkSupport: true,
                },
                rename: {
                    dynamicRegistration: true,
                    prepareSupport: true,
                },
            },
            workspace: {
                didChangeConfiguration: {
                    dynamicRegistration: true,
                },
            },
        };
        const defaultOptions = {
            capabilities: this.clientCapabilities
                ? typeof this.clientCapabilities === "function"
                    ? this.clientCapabilities(defaultClientCapabilities)
                    : this.clientCapabilities
                : defaultClientCapabilities,
            initializationOptions: this.initializationOptions,
            processId: null,
            rootUri: this.rootUri,
            workspaceFolders: this.workspaceFolders,
        };
        return defaultOptions;
    }
    async initialize() {
        const { capabilities } = await this.request("initialize", this.getInitializationOptions(), timeout * 3);
        this.capabilities = capabilities;
        this.notify("initialized", {});
        this.ready = true;
    }
    close() {
        this.client.close();
    }
    textDocumentDidOpen(params) {
        return this.notify("textDocument/didOpen", params);
    }
    textDocumentDidChange(params) {
        return this.notify("textDocument/didChange", params);
    }
    async textDocumentHover(params) {
        return await this.request("textDocument/hover", params, timeout);
    }
    async textDocumentCompletion(params) {
        return await this.request("textDocument/completion", params, timeout);
    }
    async completionItemResolve(item) {
        return await this.request("completionItem/resolve", item, timeout);
    }
    async textDocumentDefinition(params) {
        return await this.request("textDocument/definition", params, timeout);
    }
    async textDocumentCodeAction(params) {
        return await this.request("textDocument/codeAction", params, timeout);
    }
    async textDocumentRename(params) {
        return await this.request("textDocument/rename", params, timeout);
    }
    async textDocumentPrepareRename(params) {
        return await this.request("textDocument/prepareRename", params, timeout);
    }
    attachPlugin(plugin) {
        this.plugins.push(plugin);
    }
    detachPlugin(plugin) {
        const i = this.plugins.indexOf(plugin);
        if (i === -1) {
            return;
        }
        this.plugins.splice(i, 1);
        if (this.autoClose) {
            this.close();
        }
    }
    request(method, params, timeout) {
        return this.client.request({ method, params }, timeout);
    }
    notify(method, params) {
        return this.client.notify({ method, params });
    }
    processNotification(notification) {
        for (const plugin of this.plugins) {
            plugin.processNotification(notification);
        }
    }
}
class LanguageServerPlugin {
    view;
    allowHTMLContent;
    client;
    documentUri;
    languageId;
    documentVersion;
    changesTimeout;
    constructor(view, allowHTMLContent = false) {
        this.view = view;
        this.allowHTMLContent = allowHTMLContent;
        this.client = this.view.state.facet(client);
        this.documentUri = this.view.state.facet(documentUri);
        this.languageId = this.view.state.facet(languageId);
        this.documentVersion = 0;
        this.changesTimeout = 0;
        this.client.attachPlugin(this);
        this.initialize({
            documentText: this.view.state.doc.toString(),
        });
    }
    update({ docChanged }) {
        if (!docChanged) {
            return;
        }
        if (this.changesTimeout) {
            clearTimeout(this.changesTimeout);
        }
        this.changesTimeout = self.setTimeout(() => {
            this.sendChange({
                documentText: this.view.state.doc.toString(),
            });
        }, changesDelay);
    }
    destroy() {
        this.client.detachPlugin(this);
    }
    async initialize({ documentText }) {
        if (this.client.initializePromise) {
            await this.client.initializePromise;
        }
        this.client.textDocumentDidOpen({
            textDocument: {
                uri: this.documentUri,
                languageId: this.languageId,
                text: documentText,
                version: this.documentVersion,
            },
        });
    }
    async sendChange({ documentText }) {
        if (!this.client.ready) {
            return;
        }
        try {
            await this.client.textDocumentDidChange({
                textDocument: {
                    uri: this.documentUri,
                    version: this.documentVersion++,
                },
                contentChanges: [{ text: documentText }],
            });
        }
        catch (e) {
            console.error(e);
        }
    }
    requestDiagnostics(view) {
        this.sendChange({ documentText: view.state.doc.toString() });
    }
    async requestHoverTooltip(view, { line, character }) {
        if (!(this.client.ready && this.client.capabilities?.hoverProvider)) {
            return null;
        }
        this.sendChange({ documentText: view.state.doc.toString() });
        const result = await this.client.textDocumentHover({
            textDocument: { uri: this.documentUri },
            position: { line, character },
        });
        if (!result) {
            return null;
        }
        const { contents, range } = result;
        let pos = posToOffset(view.state.doc, { line, character });
        let end;
        if (range) {
            pos = posToOffset(view.state.doc, range.start);
            end = posToOffset(view.state.doc, range.end);
        }
        if (pos == null) {
            return null;
        }
        const dom = document.createElement("div");
        dom.classList.add("documentation");
        if (this.allowHTMLContent) {
            dom.innerHTML = formatContents(contents);
        }
        else {
            dom.textContent = formatContents(contents);
        }
        return {
            pos,
            end,
            create: (_view) => ({ dom }),
            above: true,
        };
    }
    async requestCompletion(context, { line, character }, { triggerKind, triggerCharacter, }) {
        if (!(this.client.ready && this.client.capabilities?.completionProvider)) {
            return null;
        }
        this.sendChange({
            documentText: context.state.doc.toString(),
        });
        const result = await this.client.textDocumentCompletion({
            textDocument: { uri: this.documentUri },
            position: { line, character },
            context: {
                triggerKind,
                triggerCharacter,
            },
        });
        if (!result) {
            return null;
        }
        let items = "items" in result ? result.items : result;
        const [_span, match] = prefixMatch(items);
        if (!match) {
            return null;
        }
        const token = context.matchBefore(match);
        let { pos } = context;
        if (token) {
            pos = token.from;
            const word = token.text.toLowerCase();
            if (/^\w+$/.test(word)) {
                items = items
                    .filter(({ label, filterText }) => {
                    const text = filterText ?? label;
                    return text.toLowerCase().startsWith(word);
                })
                    .sort((a, b) => {
                    const aText = a.sortText ?? a.label;
                    const bText = b.sortText ?? b.label;
                    switch (true) {
                        case aText.startsWith(token.text) &&
                            !bText.startsWith(token.text):
                            return -1;
                        case !aText.startsWith(token.text) &&
                            bText.startsWith(token.text):
                            return 1;
                    }
                    return 0;
                });
            }
        }
        const options = items.map((item) => {
            const { detail, labelDetails, label, kind, textEdit, insertText, documentation, additionalTextEdits, } = item;
            const completion = {
                label,
                detail: labelDetails?.detail || detail,
                apply(view, _completion, from, to) {
                    if (textEdit && isLSPTextEdit(textEdit)) {
                        view.dispatch(insertCompletionText(view.state, textEdit.newText, posToOffsetOrZero(view.state.doc, textEdit.range.start), posToOffsetOrZero(view.state.doc, textEdit.range.end)));
                    }
                    else {
                        view.dispatch(insertCompletionText(view.state, 
                        // Prefer insertText, otherwise fallback to label
                        insertText || label, from, to));
                    }
                    if (!additionalTextEdits) {
                        return;
                    }
                    const sortedEdits = additionalTextEdits.sort(({ range: { end: a } }, { range: { end: b } }) => {
                        if (posToOffsetOrZero(view.state.doc, a) <
                            posToOffsetOrZero(view.state.doc, b)) {
                            return 1;
                        }
                        if (posToOffsetOrZero(view.state.doc, a) >
                            posToOffsetOrZero(view.state.doc, b)) {
                            return -1;
                        }
                        return 0;
                    });
                    for (const textEdit of sortedEdits) {
                        view.dispatch(view.state.update({
                            changes: {
                                from: posToOffsetOrZero(view.state.doc, textEdit.range.start),
                                to: posToOffset(view.state.doc, textEdit.range.end),
                                insert: textEdit.newText,
                            },
                        }));
                    }
                },
                type: kind && CompletionItemKindMap[kind].toLowerCase(),
            };
            // Support lazy loading of documentation through completionItem/resolve
            if (this.client.capabilities?.completionProvider?.resolveProvider) {
                completion.info = async () => {
                    try {
                        const resolved = await this.client.completionItemResolve(item);
                        const dom = document.createElement("div");
                        dom.classList.add("documentation");
                        const content = resolved.documentation || documentation;
                        if (this.allowHTMLContent) {
                            dom.innerHTML = formatContents(content);
                        }
                        else {
                            dom.textContent = formatContents(content);
                        }
                        return dom;
                    }
                    catch (e) {
                        console.error("Failed to resolve completion item:", e);
                        // Fallback to existing documentation if resolve fails
                        if (documentation) {
                            const dom = document.createElement("div");
                            dom.classList.add("documentation");
                            if (this.allowHTMLContent) {
                                dom.innerHTML = formatContents(documentation);
                            }
                            else {
                                dom.textContent = formatContents(documentation);
                            }
                            return dom;
                        }
                        return null;
                    }
                };
            }
            else if (documentation) {
                // Fallback for servers without resolve support
                completion.info = () => {
                    const dom = document.createElement("div");
                    dom.classList.add("documentation");
                    if (this.allowHTMLContent) {
                        dom.innerHTML = formatContents(documentation);
                    }
                    else {
                        dom.textContent = formatContents(documentation);
                    }
                    return dom;
                };
            }
            return completion;
        });
        return {
            from: pos,
            options,
            filter: false,
        };
    }
    async requestDefinition(_view, { line, character }) {
        if (!(this.client.ready && this.client.capabilities?.definitionProvider)) {
            return;
        }
        const result = await this.client.textDocumentDefinition({
            textDocument: { uri: this.documentUri },
            position: { line, character },
        });
        if (!result)
            return;
        const locations = Array.isArray(result) ? result : [result];
        if (locations.length === 0)
            return;
        // For now just handle the first location
        const location = locations[0];
        if (!location)
            return;
        const uri = "uri" in location ? location.uri : location.targetUri;
        const range = "range" in location ? location.range : location.targetRange;
        console.debug(`Definition found at ${uri}:${range.start.line}:${range.start.character}`);
        // Not from the same document
        if (uri !== this.documentUri) {
            return;
        }
        this.view.dispatch(this.view.state.update({
            selection: {
                anchor: posToOffsetOrZero(this.view.state.doc, range.start),
                head: posToOffset(this.view.state.doc, range.end),
            },
        }));
        return { uri, range };
    }
    processNotification(notification) {
        try {
            switch (notification.method) {
                case "textDocument/publishDiagnostics":
                    this.processDiagnostics(notification.params);
            }
        }
        catch (error) {
            logger(error);
        }
    }
    async processDiagnostics(params) {
        if (params.uri !== this.documentUri) {
            return;
        }
        const diagnostics = params.diagnostics.map(async ({ range, message, severity, code }) => {
            const actions = await this.requestCodeActions(range, [
                code,
            ]);
            return {
                from: posToOffsetOrZero(this.view.state.doc, range.start),
                to: posToOffsetOrZero(this.view.state.doc, range.end),
                severity: {
                    [DiagnosticSeverity.Error]: "error",
                    [DiagnosticSeverity.Warning]: "warning",
                    [DiagnosticSeverity.Information]: "info",
                    [DiagnosticSeverity.Hint]: "info",
                }[severity ?? DiagnosticSeverity.Error],
                message,
                actions: actions?.map((action) => ({
                    name: "command" in action &&
                        typeof action.command === "object"
                        ? action.command?.title || action.title
                        : action.title,
                    apply: async () => {
                        if ("edit" in action && action.edit) {
                            // Apply workspace edit
                            for (const change of action.edit.changes?.[this.documentUri] || []) {
                                this.view.dispatch(this.view.state.update({
                                    changes: {
                                        from: posToOffsetOrZero(this.view.state.doc, change.range.start),
                                        to: posToOffset(this.view.state.doc, change.range.end),
                                        insert: change.newText,
                                    },
                                }));
                            }
                        }
                        if ("command" in action && action.command) {
                            // TODO: Implement command execution
                            // Execute command if present
                            logger("Executing command:", action.command);
                        }
                    },
                })) || [],
            };
        });
        const resolvedDiagnostics = await Promise.all(diagnostics);
        this.view.dispatch(setDiagnostics(this.view.state, resolvedDiagnostics));
    }
    async requestCodeActions(range, diagnosticCodes) {
        if (!(this.client.ready && this.client.capabilities?.codeActionProvider)) {
            return null;
        }
        return await this.client.textDocumentCodeAction({
            textDocument: { uri: this.documentUri },
            range,
            context: {
                diagnostics: [
                    {
                        range,
                        code: diagnosticCodes[0],
                        source: this.languageId,
                        message: "",
                    },
                ],
            },
        });
    }
    async requestRename(view, { line, character }) {
        if (!this.client.ready) {
            showErrorMessage(view, "Language server not ready");
            return;
        }
        if (!this.client.capabilities?.renameProvider) {
            showErrorMessage(view, "Rename not supported by language server");
            return;
        }
        try {
            // First check if rename is possible at this position
            const prepareResult = await this.client.textDocumentPrepareRename({
                textDocument: { uri: this.documentUri },
                position: { line, character },
            });
            if (!prepareResult || "defaultBehavior" in prepareResult) {
                showErrorMessage(view, "Cannot rename this symbol");
                return;
            }
            // Create popup input
            const popup = document.createElement("div");
            popup.className = "cm-rename-popup";
            popup.style.cssText =
                "position: absolute; padding: 4px; background: white; border: 1px solid #ddd; box-shadow: 0 2px 8px rgba(0,0,0,.15); z-index: 99;";
            const input = document.createElement("input");
            input.type = "text";
            input.style.cssText =
                "width: 200px; padding: 4px; border: 1px solid #ddd;";
            // Get current word as default value
            const range = "range" in prepareResult ? prepareResult.range : prepareResult;
            const from = posToOffset(view.state.doc, range.start);
            if (from == null) {
                return;
            }
            const to = posToOffset(view.state.doc, range.end);
            input.value = view.state.doc.sliceString(from, to);
            popup.appendChild(input);
            // Position the popup near the word
            const coords = view.coordsAtPos(from);
            if (!coords)
                return;
            popup.style.left = `${coords.left}px`;
            popup.style.top = `${coords.bottom + 5}px`;
            // Handle input
            const handleRename = async () => {
                const newName = input.value.trim();
                if (!newName) {
                    showErrorMessage(view, "New name cannot be empty");
                    popup.remove();
                    return;
                }
                if (newName === input.defaultValue) {
                    popup.remove();
                    return;
                }
                try {
                    const edit = await this.client.textDocumentRename({
                        textDocument: { uri: this.documentUri },
                        position: { line, character },
                        newName,
                    });
                    if (!edit?.changes) {
                        showErrorMessage(view, "No changes to apply");
                        popup.remove();
                        return;
                    }
                    // Apply all changes
                    for (const [uri, changes] of Object.entries(edit.changes)) {
                        if (uri !== this.documentUri) {
                            showErrorMessage(view, "Multi-file rename not supported yet");
                            continue;
                        }
                        // Sort changes in reverse order to avoid position shifts
                        const sortedChanges = changes.sort((a, b) => {
                            const posA = posToOffset(view.state.doc, a.range.start);
                            const posB = posToOffset(view.state.doc, b.range.start);
                            return (posB ?? 0) - (posA ?? 0);
                        });
                        view.dispatch(view.state.update({
                            changes: sortedChanges.map((change) => ({
                                from: posToOffset(view.state.doc, change.range.start) ?? 0,
                                to: posToOffset(view.state.doc, change.range.end) ?? 0,
                                insert: change.newText,
                            })),
                        }));
                    }
                }
                catch (error) {
                    showErrorMessage(view, `Rename failed: ${error instanceof Error ? error.message : "Unknown error"}`);
                }
                finally {
                    popup.remove();
                }
            };
            input.addEventListener("keydown", (e) => {
                if (e.key === "Enter") {
                    handleRename();
                }
                else if (e.key === "Escape") {
                    popup.remove();
                }
                e.stopPropagation(); // Prevent editor handling
            });
            // Handle clicks outside
            const handleOutsideClick = (e) => {
                if (!popup.contains(e.target)) {
                    popup.remove();
                    document.removeEventListener("mousedown", handleOutsideClick);
                }
            };
            document.addEventListener("mousedown", handleOutsideClick);
            // Add to DOM
            document.body.appendChild(popup);
            input.focus();
            input.select();
        }
        catch (error) {
            showErrorMessage(view, `Rename failed: ${error instanceof Error ? error.message : "Unknown error"}`);
        }
    }
}
export function languageServer(options) {
    const { serverUri, ...rest } = options;
    return languageServerWithTransport({
        ...rest,
        transport: new WebSocketTransport(serverUri),
    });
}
export function languageServerWithTransport(options) {
    let plugin = null;
    const shortcuts = {
        rename: "F2",
        goToDefinition: "ctrlcmd", // ctrlcmd means Ctrl on Windows/Linux, Cmd on Mac
        ...options.keyboardShortcuts,
    };
    return [
        client.of(options.client ||
            new LanguageServerClient({ ...options, autoClose: true })),
        documentUri.of(options.documentUri),
        languageId.of(options.languageId),
        ViewPlugin.define((view) => {
            plugin = new LanguageServerPlugin(view, options.allowHTMLContent);
            return plugin;
        }),
        hoverTooltip((view, pos) => plugin?.requestHoverTooltip(view, offsetToPos(view.state.doc, pos)) ?? null),
        autocompletion({
            override: [
                async (context) => {
                    if (plugin == null) {
                        return null;
                    }
                    const { state, pos, explicit } = context;
                    const line = state.doc.lineAt(pos);
                    let trigKind = CompletionTriggerKind.Invoked;
                    let trigChar;
                    if (!explicit &&
                        plugin.client.capabilities?.completionProvider?.triggerCharacters?.includes(line.text[pos - line.from - 1] || "")) {
                        trigKind = CompletionTriggerKind.TriggerCharacter;
                        trigChar = line.text[pos - line.from - 1];
                    }
                    if (trigKind === CompletionTriggerKind.Invoked &&
                        !context.matchBefore(/\w+$/)) {
                        return null;
                    }
                    return await plugin.requestCompletion(context, offsetToPos(state.doc, pos), {
                        triggerCharacter: trigChar,
                        triggerKind: trigKind,
                    });
                },
            ],
        }),
        EditorView.domEventHandlers({
            click: (event, view) => {
                if (shortcuts.goToDefinition === "ctrlcmd" &&
                    (event.ctrlKey || event.metaKey)) {
                    const pos = view.posAtCoords({
                        x: event.clientX,
                        y: event.clientY,
                    });
                    if (pos && plugin) {
                        plugin
                            .requestDefinition(view, offsetToPos(view.state.doc, pos))
                            .catch((error) => showErrorMessage(view, `Go to definition failed: ${error instanceof Error ? error.message : "Unknown error"}`));
                        event.preventDefault();
                    }
                }
            },
            keydown: (event, view) => {
                if (event.key === shortcuts.rename && plugin) {
                    const pos = view.state.selection.main.head;
                    plugin.requestRename(view, offsetToPos(view.state.doc, pos));
                    event.preventDefault();
                }
                else if (shortcuts.goToDefinition !== "ctrlcmd" &&
                    event.key === shortcuts.goToDefinition &&
                    plugin) {
                    const pos = view.state.selection.main.head;
                    plugin
                        .requestDefinition(view, offsetToPos(view.state.doc, pos))
                        .catch((error) => showErrorMessage(view, `Go to definition failed: ${error instanceof Error ? error.message : "Unknown error"}`));
                    event.preventDefault();
                }
            },
        }),
    ];
}
//# sourceMappingURL=index.js.map